"""Configuration file"""
pages = ['Home', 'General', 'Analysis', 'Statistics', 'Parsing', 'Plotting']

'''General settings'''
General = {
    'run_analysis': {'type': 'CheckButton', 'Recommended': 'true',
                     'status': 'important',
                     'Description': 'true or false. Can skip this step if json files have already been created.',
                     'label': 'Run analysis'},

    'run_statistics': {'type': 'CheckButton', 'Recommended': 'true',
                       'status': 'important',
                       'Description': 'true or false.', 'label': 'Run statistics'},

    'run_plotting': {'type': 'CheckButton', 'Recommended': 'true',
                     'status': 'important',
                     'Description': 'true or false.', 'label': 'Run plotting'},

    'verbose': {'type': 'CheckButton', 'Recommended': 'true',
                'Description': 'true or false.', 'label': 'Verbose fRAT stages'},

    'verbose_cmd_line_args': {'type': 'CheckButton', 'Recommended': 'false',
                              'Description': 'true or false.', 'label': 'Verbose command line arguments'},

    'multicore_processing': {'type': 'CheckButton', 'Recommended': 'true',
                             'Description': 'true or false. Use multicore processing during analysis? Multicore processing currently works within participants not between them. Recommended: true'},

    'max_core_usage': {'type': 'OptionMenu', 'Recommended': 'max', 'Options': ['max', 6, 5, 4, 3, 2, 1],
                       'save_as': 'string',
                       'Description': "'max' to select number of cores available on the system, alternatively an int to manually select number of cores to use. Recommended: 'max'"},

    'brain_file_loc': {'type': 'Entry', 'Recommended': "", 'save_as': 'string',
                       'label': 'Base directory of subjects',
                       'Description': 'Either the absolute location of brain files or blank, if blank then a browser window will allow you to search for the files at runtime. If passing in this information as a command line flag, this will be ignored.'},

    'report_output_folder': {'type': 'Entry', 'Recommended': "", 'save_as': 'string',
                             'label': 'fRAT output directory location',
                             'Description': 'Either the absolute location of json files or blank, if blank then a browser window will allow you to search for the files at runtime. If passing in this information as a command line flag, this will be ignored.'},

    'averaging_type': {'type': 'OptionMenu',
                       'Recommended': 'Participant averaged',
                       'Options': ['Session averaged', 'Participant averaged'],
                       'save_as': 'string',
                       'Description': 'Participant averaged or Session averaged.\n'
                                      'This setting is used to determine which statistics to use for plotting, '
                                      'and when accessing results (for example through the '
                                      'interactive report). \nNote: Histograms will always use the raw results. The '
                                      'linear mixed model from the statistics will always use session averaged data'},

    'parameter_file': {'type': 'Entry', 'Recommended': "paramValues.csv", 'save_as': 'string',
                       'Description': 'Recommended: paramValues.csv\n'
                                      'Name of the file to parse for critical params. Option added to allow quick '
                                      'swapping between different parameter files.'},

    'file_cleanup': {'type': 'OptionMenu', 'Recommended': 'move', 'Options': ['move', 'delete'], 'save_as': 'string',
                     'Description': 'Move or delete intermediate files.', 'label': 'File cleanup method'},

    'Installation testing': {'type': 'subheading'},

    'run_tests': {'type': 'Button', 'Command': 'run_tests', 'Text': 'Run installation tests', 'Pass self': True,
                  'Description': 'Requires example data set. The link can be found on the fRAT website.'},

    'delete_test_folder': {'type': 'OptionMenu',
                           'Recommended': 'If completed without error',
                           'Options': ['Always', 'If completed without error', 'Never'],
                           'save_as': 'string',
                           'Description': 'Option to choose whether the folder generated while running tests is '
                                          'deleted upon completion. This only applies when running the full comparison.'},

    'verbose_errors': {'type': 'CheckButton', 'Recommended': 'true',
                       'Description': 'true or false. '
                                      'Print all missing files and differences found during testing to the terminal.'}
}

'''Analysis settings'''
Analysis = {
    'atlas_number': {'type': 'OptionMenu', 'Recommended': 'HarvardOxford-cort', 'save_as': 'string', 'label': 'Atlas',
                     'Options': [
                         'Cerebellum-MNIflirt',
                         'Cerebellum-MNIfnirt',
                         'HarvardOxford-cort',
                         'HarvardOxford-sub',
                         'JHU-ICBM-labels',
                         'JHU-ICBM-tracts',
                         'juelich',
                         'MNI',
                         'SMATT-labels',
                         'STN',
                         'striatum-structural',
                         'Talairach-labels',
                         'Thalamus'
                     ],
                     'Description': ''},

    'input_folder_name': {'type': 'Entry', 'Recommended': "func_cleaned", 'save_as': 'string',
                          'Description': 'Folder found in each subjects directory containing the files to be analysed. '
                                         'func_cleaned is the default option as this folder will automatically be '
                                         'created when making statmaps. If the "Noise volume included in time series" '
                                         'option was set to true, or motion outlier removal was used when creating '
                                         'the statmaps, this folder will contain cleaned versions of the original func '
                                         'files. However if these options were not used when creating the statmaps, '
                                         'the folder will still be present, however the files will be identical to '
                                         'those in the "func" folder.'},

    'output_folder': {'type': 'Entry', 'Recommended': 'DEFAULT', 'save_as': 'string', 'label': 'Output directory',
                      'Description': 'Directory to save output. If set to DEFAULT, output directory will be set to '
                                     'the cortical atlas used appended with "_ROI_report". '
                                     '\nExample: HarvardOxford-Cortical_ROI_report/'},

    'dof': {'type': 'Entry', 'Recommended': 12, 'label': 'DOF',
            'Description': 'Degrees of freedom for FLIRT (only used for the fMRI to anatomical alignment when using '
                           'Correlation Ratio cost function). Recommended: 12'},

    'anat_align_cost_function': {'type': 'OptionMenu',
                                 'label': 'fMRI to anatomical cost function',
                                 'Recommended': 'BBR',
                                 'Options': ['BBR', 'Correlation Ratio'],
                                 'save_as': 'string',
                                 'Description': 'BBR or Correlation Ratio. Recommended: BBR.\n'
                                                'Using BBR (Boundary-Based Registration) requires an FSL FAST '
                                                'segmentation (this will be automatically created if necessary if '
                                                'the Run FSL FAST option is set to "Run if files not found") '
                                                'and a wholehead non-brain extracted anatomical placed in the anat '
                                                'folder.'},

    'grey_matter_segment': {'type': 'CheckButton',
                            'Recommended': 'true', 'label': 'Use FAST to only include grey matter',
                            'Description': 'true or false. Recommended: true if using a cortical atlas.'
                                           '\nNote: FSL FAST segmentation files should be placed in the '
                                           'sub-{id}/fslfast/ directory. Only the FSL FAST file appended with pve_1 '
                                           'needs to be in this directory, however if all files output by FAST are '
                                           'placed in this directory, then fRAT will find the necessary file.'},

    'run_fsl_fast': {'type': 'OptionMenu',
                     'Options': ['Run if files not found', 'Never run'],
                     'Recommended': 'Run if files not found',
                     'Description': 'Recommended: "Run if files not found".\nThese files will only be searched for '
                                    '(and thus created) if "Use FSL FAST segmentation" is set to true.',
                     'label': 'Run FSL FAST',
                     'save_as': 'string'},

    'fslfast_min_prob': {'type': 'Scale', 'Recommended': 0.1, 'From': 0, 'To': 1, 'Resolution': 0.05,
                         'label': 'FSL FAST minimum probability', 'Description': 'Recommended: 0.1'},

    'stat_map_folder': {'type': 'Entry', 'Recommended': '', 'save_as': 'string',
                        'status': 'important',
                        'label': 'Statistical map folder',
                        'Description': 'Folder name which contains the statistical map files. '
                                       'Example: temporalSNR_report'},

    'stat_map_suffix': {'type': 'Entry', 'Recommended': '_tSNR.nii.gz', 'save_as': 'string',
                        'status': 'important',
                        'label': 'Statistical map suffix',
                        'Description': 'File name suffix of the statistical map files. Include the file extension. '
                                       'Example: _tSNR.img'},

    # 'bootstrap': {'type': 'CheckButton', 'Recommended': 'false',
    #               'Description': 'true or false. Calculate bootstrapped mean and confidence intervals using 10,000 iterations'},

    'conf_level_number': {'type': 'OptionMenu', 'Recommended': '95%, 1.96',
                          'Options': ['80%, 1.28', '85%, 1.44', '90%, 1.64', '95%, 1.96', '98%, 2.33', '99%, 2.58'],
                          'save_as': 'string', 'label': 'Confidence level',
                          'Description': 'Set the confidence level for confidence interval calculations.\n'
                                         'Numbers represent the confidence level and the corresponding critical z value.\n'
                                         'Recommended: 95%, 1.96.'},

    'binary_params': {'type': 'Dynamic', 'Recommended': [''], 'Options': 'Parsing["parameter_dict1"]',
                      'subtype': 'Checkbutton', 'label': 'Binary parameters',
                      'save_as': 'list', 'Description': 'Add parameters here which will either be on or off.'},

    'Outlier detection': {'type': 'subheading'},

    'noise_cutoff': {'type': 'CheckButton', 'Recommended': 'true',
                     'Description': 'true or false. Calculate a minimum cutoff value to be included in an ROI,'
                                    'based on voxels not assigned an ROI or that have been excluded from analysis. '
                                    'Voxels with values of 0 are not included when calculating the noise cutoff. \n'
                                    'Useful for statistical maps where extracranial voxels are likely to have much '
                                    'lower values than those inside the brain such as tSNR maps. Recommended: true.'},

    'gaussian_outlier_detection': {'type': 'CheckButton', 'Recommended': 'true',
                                   'label': 'Gaussian outlier (GauO) detection',
                                   'Description': 'true or false. Fit a gaussian to the data to determine outliers using Elliptic Envelope. '
                                                  'Recommended: true.'},

    'gaussian_outlier_contamination': {'type': 'Scale', 'Recommended': 0.1, 'From': 0, 'To': 1, 'Resolution': 0.01,
                                       'label': 'GauO contamination percentage',
                                       'Description': 'Percent of expected outliers in dataset\n'
                                                      f'Recommended: 0.1'},

    'gaussian_outlier_location': {'type': 'OptionMenu', 'Recommended': 'below gaussian',
                                  'Options': ['below gaussian', 'above gaussian', 'both'],
                                  'save_as': 'string', 'label': 'GauO location',
                                  'Description': 'Data to remove (if gaussian outlier detection is true).\n'
                                                 'For example: if set to below gaussian, data below the gaussian will be removed.\n'
                                                 'Recommended: below gaussian.'},
}

'''Statistics settings'''
Statistics = {
    'create_statistics_options_file': {'type': 'Button', 'Command': 'create_statistics_file', 'Pass self': 'false',
                                       'Text': 'Create statisticsOptions.csv',
                                       'Description': 'Create file in base folder to choose statistics options. Such as '
                                                      'which parameters to exclude from analysis and what main effect '
                                                      'contrasts to calculate.'},

    'automatically_create_statistics_options_file': {'type': 'CheckButton',
                                                     'Recommended': 'true',
                                                     'Description': 'true or false.\n'
                                                                    "Usually statisticsOptions.csv is automatically created when creating "
                                                                    "paramValues.csv. Deselect this option if you won't be running "
                                                                    "the statistics step. The create statisticsOptions.csv button "
                                                                    "above can also be used to manually create this file."},

    'statistics_subfolder_name': {'type': 'Entry',
                                  'Recommended': 'stats',
                                  'save_as': 'string',
                                  'Description': 'Directory name for statistics folder.'},

    'print_result': {'type': 'CheckButton', 'Recommended': 'true',
                     'label': 'Print results to terminal',
                     'Description': 'true or false. Prints results to terminal if true in addition to saving results '
                                    'to file.'},

    'minimum_voxels': {'type': 'Entry',
                       'Recommended': 400,
                       'Description': 'For bootstrapped change versus baseline, for each ROI, the average number of '
                                      'voxels per session for an ROI must be above this value to be included in the '
                                      'analysis.'
                                      'For running the linear mixed model, for each ROI, any sessions with a voxel '
                                      'count below this value will be removed.'
                                      '\nHighly recommended to set a value here, as ROIs with a small number of voxels '
                                      'may suggest poor fitting.'
                                      '\nRecommended value 400'},

    'bootstrap_samples': {'type': 'Entry', 'Recommended': 1000,
                          'Description': 'Recommended value 1000. \n'
                                         'Note: Bootstrapping is only used to calculate percentage change versus '
                                         'baseline.'},

    'bootstrap_confidence_interval': {'type': 'Entry', 'Recommended': 95,
                                      'Description': 'Recommended value: 95 \n'
                                                     'Note: Bootstrapping is only used to calculate percentage change '
                                                     'versus baseline.'},

    'regional_stats_rois': {'type': 'Entry', 'Recommended': 'all', 'save_as': 'list',
                            'label': 'ROIs to calculate statistics for',
                            'Description': "Provide a comma-separated list of regions, e.g. '3, 5', the string "
                                           "'all' for all rois or the string 'Runtime' to provide regions at runtime."},

    'include_as_variable': {'type': 'Dynamic', 'Recommended': 'INCLUDE ALL VARIABLES',
                            'status': 'important',
                            'Options': 'Parsing["parameter_dict1"]',
                            'subtype': 'Checkbutton',
                            'save_as': 'list',
                            'Description': 'Select which variables to include in statistical analysis.\nUsed to '
                                           'determine which variables to use as fixed effects in linear mixed models '
                                           'and which variables to take into account when balancing data for the main '
                                           'effect t test data.'},

    'brain_map_p_thresh': {'type': 'Entry', 'Recommended': 0.05,
                           'label': 'Coefficient map p-threshold',
                           'Description': "P-value threshold to use when creating brain coefficient maps. "
                                          "Any fixed effect that doesn't have a p-value equal to or less than this "
                                          "value will not be included in the coefficient map.\n",
                           'status': 'important'},

    'T-tests': {'type': 'subheading'},

    'run_t_tests': {'type': 'CheckButton', 'status': 'important',
                    'Recommended': 'true',
                    'Description': 'true or false.\n'},

    'IV_type': {'type': 'Dynamic', 'Recommended': 'FILL IV TYPE AS BETWEEN-SUBJECTS',
                'Options': 'Parsing["parameter_dict1"]',
                'subtype': 'OptionMenu2', 'save_as': 'list', 'Options2': ['Within-subjects', 'Between-subjects'],
                'label': 'IV type', 'status': 'important', 'DefaultNumber': 1,
                'Description': 'Type of variable collected. Used to choose which t-test to use for pairwise '
                               'comparisons.'},

    # 'glm_statistic': {'type': 'OptionMenu',
    #                   'label': 'GLM statistic for overall effect',
    #                   'Recommended': 'Mean',
    #                   'Options': ['Mean', 'Median'],
    #                   'save_as': 'string',
    #                   'Description': 'Mean or Median.\nNote: this setting is only used when '
    #                                  'looking at the overall effect, as statistics ran on individual ROIs will use raw '
    #                                  'voxel values.'},

    'Linear mixed models': {'type': 'subheading'},

    'run_linear_mixed_models': {'type': 'CheckButton', 'status': 'important',
                                'Recommended': 'true',
                                'Description': 'true or false.\n'},

    'categorical_variables': {'type': 'Dynamic', 'Recommended': [''], 'Options': 'Parsing["parameter_dict1"]',
                              'subtype': 'Checkbutton',
                              'save_as': 'list',
                              'Description': 'Select which variables (if any) are categorical. Used for the LMM.'},

    'main_effects': {'type': 'CheckButton',
                     'Recommended': 'true',
                     'label': 'Compute main effects',
                     'Description': 'true or false.\n'
                                    'Note: This option is independent from the other effect calculations.'},

    'main_and_interaction_effects': {'type': 'CheckButton',
                                     'Recommended': 'true',
                                     'label': 'Compute main and interaction effects',
                                     'Description': 'true or false.\n'
                                                    'Note: This option is independent from the other effect calculations.'},

    'interaction_effects': {'type': 'CheckButton',
                            'Recommended': 'false',
                            'label': 'Compute interaction effects',
                            'Description': 'true or false.\n'
                                           'Note: This option is independent from the other effect calculations.'},

    'R2 vs voxel count LMM': {'type': 'subheading'},

    'max_below_thresh': {'type': 'Entry', 'Recommended': 0, 'label': 'Maximum percent of sessions below thresh',
                             'Description': 'Recommended value 0. \n'
                                            'For a given ROI, this value sets the maximum percent of sessions that can '
                                            'be excluded (due to having insufficient voxel count) before the ROI is not '
                                            'included in r2 vs voxel count statistics.'
                                            '\nWith the default value of 100(%) an ROI will not be included in this '
                                            'calculation if any sessions have been excluded.'},
}

'''Parsing settings'''
Parsing = {
    'parameter_dict1': {'type': 'Entry',
                        'status': 'important',
                        'Recommended': 'MB, SENSE', 'save_as': 'list', 'label': 'Critical parameters',
                        'Description': 'Comma-separated list of independent variables.'
                                       '\nThe critical parameter settings are used to supply the names and file name '
                                       'abbreviations of the independent variables, therefore `fRAT` supports the use '
                                       'of any parameters (and any number of them).'
                                       '\nAs these critical parameters will also be used when labelling the rows and '
                                       'columns of both the violin plots and histograms, they should be written as '
                                       'you want them to appear in these figures.'
                                       '\nNote: Leave blank if you do not want to compare between different conditions, '
                                       'for example, if you wish to see the overall tSNR for each region across the '
                                       'entire dataset.'},

    'parameter_dict2': {'type': 'Entry',
                        'status': 'important',
                        'Recommended': 'mb, s', 'save_as': 'list',
                        'label': 'Critical parameter abbreviation',
                        'Description': 'Comma-separated list of terms to parse the file name for. Each entry '
                                       'corresponds to a critical parameter above. \nOptional if using table parameter '
                                       'verification, however if the file name contains this information it can use '
                                       'this information to auto-detect the critical parameters used for each fMRI '
                                       'volume.'
                                       '\nNote: This field can be blank.'},

    'make_folder_structure': {'type': 'CheckButton', 'Recommended': 'true',
                              'Description': 'true or false. Make folder structure when creating paramValues.csv'},

    'parsing_folder': {'type': 'Entry', 'Recommended': "func", 'save_as': 'string',
                       'Description': 'Folder to find files to add to paramValues.csv. If using "Make folder '
                                      'structure" option, this will be the directory the files in the participant '
                                      'folder will be moved to.'},
}

'''General plot settings'''
Plotting = {
    'General plot settings': {'type': 'subheading'},

    'plot_dpi': {'type': 'Entry', 'Recommended': 600, 'label': 'Figure DPI',
                 'Description': 'Recommended value 600'},

    'plot_font_size': {'type': 'Entry', 'Recommended': 40, 'label': 'Figure font size',
                       'Description': 'Recommended value 30'},

    'plot_scale': {'type': 'Entry', 'Recommended': 10, 'label': 'Figure scale',
                   'Description': 'Recommended value 10'},

    'make_violin_plot': {'type': 'CheckButton', 'Recommended': 'true', 'label': 'Make violin plots',
                         'Description': 'true or false.'},

    'make_brain_table': {'type': 'CheckButton', 'Recommended': 'true', 'label': 'Make brain visualisations',
                         'Description': 'true or false.'},

    'make_one_region_fig': {'type': 'CheckButton', 'Recommended': 'true', 'label': 'Make regional barcharts',
                            'Description': 'true or false.'},

    'make_histogram': {'type': 'CheckButton', 'Recommended': 'true', 'label': 'Make regional histograms',
                       'Description': 'true or false.'},

    'colorblind_friendly_plot_colours': {'type': 'Entry', 'Recommended': '#ffeda0, #feb24c, #fc4e2a, #bd0026',
                                         'save_as': 'list',
                                         'Description': 'Hex values of colourblind friendly colour scale.'},

    'regional_fig_rois': {'type': 'Entry', 'Recommended': 'all', 'save_as': 'list', 'label': 'ROIs to plot',
                          'Description': "Provide a comma-separated list of regions to plot e.g. '3, 5', the string "
                                         "'all' for all rois or the string 'Runtime' to provide regions at runtime."},

    'Brain table': {'type': 'subheading'},

    'brain_tight_layout': {'type': 'CheckButton', 'Recommended': 'false',
                           'Description': 'true or false. Use a tight layout when laying out the figure. Recommended: false'},

    'brain_fig_value_min': {'type': 'Entry', 'Recommended': 0, 'label': 'Minimum median and mean value',
                            'Description': 'Provides the minimum value of the colourbar when creating mean and median '
                                           'images. For example, set minimum to 50 to make areas with values below 50 '
                                           'appear black.\n'
                                           'Recommended value: 0'},

    'brain_fig_value_max': {'type': 'Entry', 'Recommended': None, 'label': 'Maximum median and mean value',
                            'Description': 'Provides the maximum value of the colourbar when creating mean and median '
                                           'images. For example, set maximum to 50 to make areas with values above 50 '
                                           'appear as the brighest colour on the colourbar.\n'
                                           'Recommended value: None. Note: will default to 100 for scaled maps.'},

    'brain_x_coord': {'type': 'Entry', 'Recommended': -1,
                      'Description': 'Voxel location to slice the images at in the x axis. '
                                     'Recommended settings for both variables: 91 or 58'},

    'brain_z_coord': {'type': 'Entry', 'Recommended': 19,
                      'Description': 'Voxel location to slice the images at in the z axis. '
                                     'Recommended settings for both variables: 91 or 58'},

    'brain_table_col_labels': {'type': 'Entry', 'Recommended': 'CHANGE TO DESIRED LABEL', 'save_as': 'string',
                               'label': 'Column labels',
                               'DefaultNumber': 0, 'Description': 'Label for columns.'},

    'brain_table_row_labels': {'type': 'Entry', 'Recommended': 'CHANGE TO DESIRED LABEL', 'save_as': 'string',
                               'label': 'Row labels',
                               'DefaultNumber': 1, 'Description': 'Label for rows.'},

    'brain_table_cols': {'type': 'Dynamic', 'Recommended': 'DEFAULT', 'Options': 'Parsing["parameter_dict1"]',
                         'subtype': 'OptionMenu', 'save_as': 'string',
                         'DefaultNumber': 0, 'Description': ''},

    'brain_table_rows': {'type': 'Dynamic', 'Recommended': 'DEFAULT', 'Options': 'Parsing["parameter_dict1"]',
                         'subtype': 'OptionMenu', 'save_as': 'string',
                         'DefaultNumber': 1, 'Description': ''},

    'Violin plot': {'type': 'subheading'},

    'table_x_label': {'type': 'Entry', 'Recommended': 'tSNR mean', 'save_as': 'string',
                      'Description': ''},

    'table_y_label': {'type': 'Entry', 'Recommended': 'ROI', 'save_as': 'string',
                      'Description': ''},

    'violin_show_data': {'type': 'CheckButton', 'Recommended': 'true', 'label': 'Show data points',
                         'Description': 'true or false.'},

    'violin_jitter': {'type': 'CheckButton', 'Recommended': 'true', 'label': 'Jitter data points',
                      'Description': 'true or false.'},

    'violin_colour': {'type': 'Dynamic', 'Recommended': '#fc4e2a',
                      'Options': 'Plotting["colorblind_friendly_plot_colours"]',
                      'subtype': 'OptionMenu', 'save_as': 'string',
                      'Description': 'Hex value of colour blind friendly colour. Value taken from colorblind friendly plot colours.'},

    'boxplot_colour': {'type': 'Dynamic', 'Recommended': '#feb24c',
                       'Options': 'Plotting["colorblind_friendly_plot_colours"]',
                       'subtype': 'OptionMenu', 'save_as': 'string',
                       'Description': 'Hex value of colour blind friendly colour. Value taken from colorblind friendly plot colours.'},

    'table_cols': {'type': 'Dynamic', 'Recommended': 'DEFAULT', 'Options': 'Parsing["parameter_dict1"]',
                   'subtype': 'OptionMenu', 'save_as': 'string', 'DefaultNumber': 0,
                   'Description': ''},

    'table_rows': {'type': 'Dynamic', 'Recommended': 'DEFAULT', 'Options': 'Parsing["parameter_dict1"]',
                   'subtype': 'OptionMenu', 'save_as': 'string', 'DefaultNumber': 1,
                   'Description': ''},

    'Regional bar chart': {'type': 'subheading'},

    'single_roi_fig_label_x': {'type': 'Entry', 'Recommended': 'Multiband factor', 'save_as': 'string',
                               'Description': ""},

    'single_roi_fig_label_y': {'type': 'Entry', 'Recommended': 'temporal Signal to Noise Ratio', 'save_as': 'string',
                               'Description': ""},

    'single_roi_fig_label_fill': {'type': 'Entry', 'Recommended': 'SENSE factor', 'save_as': 'string',
                                  'Description': ""},

    'single_roi_fig_x_axis': {'type': 'Dynamic', 'Recommended': 'DEFAULT', 'Options': 'Parsing["parameter_dict1"]',
                              'subtype': 'OptionMenu', 'save_as': 'string', 'DefaultNumber': 0,
                              'Description': ''},

    'single_roi_fig_colour': {'type': 'Dynamic', 'Recommended': 'DEFAULT', 'Options': 'Parsing["parameter_dict1"]',
                              'subtype': 'OptionMenu', 'save_as': 'string', 'DefaultNumber': 1,
                              'Description': ''},

    'Regional histogram': {'type': 'subheading'},

    'histogram_binwidth': {'type': 'Entry', 'Recommended': 2, 'label': 'Bin width', 'Description': ""},

    'histogram_fig_label_x': {'type': 'Entry', 'Recommended': 'temporal Signal to Noise Ratio',
                              'save_as': 'string', 'label': 'x-axis label',
                              'Description': ''},

    'histogram_fig_label_y': {'type': 'Entry', 'Recommended': 'Frequency',
                              'save_as': 'string', 'label': 'y-axis label',
                              'Description': ''},

    'histogram_stat_line_size': {'type': 'Entry', 'Recommended': 1.5, 'label': 'Statistic line size',
                                 'Description': ''},

    'histogram_show_mean': {'type': 'CheckButton', 'Recommended': 'true', 'label': 'Show mean',
                            'Description': 'true or false.'},

    'histogram_show_median': {'type': 'CheckButton', 'Recommended': 'true', 'label': 'Show median',
                              'Description': 'true or false.'},

    'histogram_show_legend': {'type': 'CheckButton', 'Recommended': 'true', 'label': 'Show legend',
                              'Description': 'true or false.'},

    'histogram_fig_x_facet': {'type': 'Dynamic', 'Recommended': 'DEFAULT', 'label': 'x-axis facet',
                              'Options': 'Parsing["parameter_dict1"]',
                              'subtype': 'OptionMenu', 'save_as': 'string', 'DefaultNumber': 0,
                              'Description': ''},

    'histogram_fig_y_facet': {'type': 'Dynamic', 'Recommended': 'DEFAULT', 'label': 'y-axis facet',
                              'Options': 'Parsing["parameter_dict1"]',
                              'subtype': 'OptionMenu', 'save_as': 'string',
                              'DefaultNumber': 1, 'Description': ''},

    'histogram_fig_colour': {'type': 'Dynamic', 'Recommended': '#fc4e2a', 'label': 'Bin colour',
                             'Options': 'Plotting["colorblind_friendly_plot_colours"]',
                             'subtype': 'OptionMenu', 'save_as': 'string',
                             'Description': 'Hex value of colour blind friendly colour. Value taken from colorblind friendly plot colours.'}
}



import argparse
import ast
import os
import re
import shutil
import sys
from glob import glob
from pathlib import Path
from tkinter import Tk, filedialog
from types import SimpleNamespace

import multiprocess as mp
import nibabel as nib
import pandas as pd
import toml

from fRAT.utils.fRAT_config_setup import *

config = None
version = None


class Utils:
    @staticmethod
    def convert_toml_input_to_python_object(x):
        try:
            if x in ['true', 'false']:
                x = x.title()

            x = ast.literal_eval(x)

            if isinstance(x, tuple):
                x = list(x)

        except (ValueError, SyntaxError):
            pass

        return x

    @staticmethod
    def argparser():
        # Create the parser
        parser = argparse.ArgumentParser(prog='fRAT.py', usage='%(prog)s [arguments]',
                                         description='Convert voxelwise statistics to regionwise statistics for fMRI '
                                                     'data.',
                                         epilog='Supplying arguments in this way is for advanced users only. It is '
                                                'recommended that settings are changed using the GUI or by '
                                                'editing the fRAT_config.toml file. '
                                                'Arguments should be given in toml format. For example: '
                                                'true should be used instead of True and strings should be in '
                                                'quotation marks. Where a comma-separated list is supposed to be '
                                                'given, this should be in the format: ["mb", "sense"].')

        # Add the arguments
        parser.add_argument('--make_table', dest='make_table',
                            help='true or false. Use this flag to create a csv file to store parameter information '
                                 'about files. Recommended that this file is created and filled in before fRAT '
                                 'execution.')

        arg_categories = pages[1:]

        for category in arg_categories:
            for arg in eval(category):
                if eval(category)[arg]['type'] == 'Button':
                    # Skip button as they don't need help text
                    continue

                try:
                    help_text = eval(category)[arg]["Description"].replace("%", "%%")
                except KeyError:
                    continue

                if help_text == "":
                    help_text = f"Recommended value: {eval(category)[arg]['Recommended']}"

                parser.add_argument(f'--{arg}', dest=arg, help=help_text)

        # Execute the parse_args() method
        args = parser.parse_args()

        return args

    @staticmethod
    def find_files(directory, *extensions):
        files = []
        for extension in extensions:
            if extension[0] == ".":
                extension.lstrip(".")

            these_files = [os.path.basename(f) for f in glob(f"{directory}/*.{extension}")]

            if these_files:
                files.extend(these_files)

        return files

    @staticmethod
    def dataframe_to_dict(dataframe):
        roidict = {}
        for column in dataframe.columns:
            roidict[column] = dataframe[column].dropna().to_numpy().tolist()

        return roidict

    @staticmethod
    def dict_to_dataframe(roidict):
        return pd.DataFrame.from_dict(roidict, orient='index').transpose()

    @staticmethod
    def file_browser(title='', chdir=False):
        root = Tk()
        root.withdraw()  # Hide tkinter root window

        directory = filedialog.askdirectory(title=title)

        root.destroy()  # Destroy tkinter root window

        if not directory:
            raise FileNotFoundError('No folder selected.')

        if chdir:
            os.chdir(directory)

        if config is None or config.verbose:
            print(f"Selected directory: {directory}")

        return directory

    @staticmethod
    def print_and_save(file_object, print_to_terminal, *args):
        for line in args:
            if print_to_terminal:
                print(line)
            file_object.write(str(line) + "\n")

    @staticmethod
    def save_config(newdir, config_path, config_filename, additional_info=None, relevant_sections='all', new_config_name='config_log'):
        with open(f'{newdir}/{new_config_name}.toml', 'w') as f, open(f'{config_path}/{config_filename}', 'r') as r:
            f.write(f"# General information\n"
                    f"version = {version}\n"
                    f"config_file_used = '{config_filename}'\n")

            if additional_info:
                for line in additional_info:
                    f.write(line)

                f.write('\n')

            f.write('\n')

            current_section = None
            for line in r:
                if line.startswith('#') and not line.startswith('##'):
                    current_section = line.replace('#', '')[1:-1]

                if relevant_sections == 'all':
                    f.write(line)
                elif current_section in relevant_sections:
                    f.write(line)

    @staticmethod
    def move_file(old_name, original_dir, new_dir, copy=False, rename_copy=True, parameter_file=False):
        if not original_dir.endswith('/'):
            original_dir += '/'

        if not new_dir.endswith('/'):
            new_dir += '/'

        if parameter_file:
            new_name = "paramValues.csv"
        else:
            new_name = old_name

        if copy:
            if rename_copy:
                shutil.copy(f"{original_dir}{old_name}", f"{new_dir}copy_{new_name}")
            else:
                shutil.copy(f"{original_dir}{old_name}", f"{new_dir}{new_name}")
        else:
            os.rename(f"{original_dir}{old_name}", f"{new_dir}{new_name}")

    @staticmethod
    def check_and_make_dir(path, delete_old=False):
        if delete_old and os.path.exists(path):
            shutil.rmtree(path)
            Utils.mk_dir(path)

        elif not os.path.exists(path):
            Utils.mk_dir(path)

    @staticmethod
    def find_chosen_rois(all_rois, func_name, config_region_var):
        if config_region_var == 'Runtime':  # If no ROI has been selected
            chosen_rois = []
            print("\n")

            for roi_num, roi in enumerate(all_rois):
                print("{roi_num}: {roi}".format(roi_num=roi_num, roi=roi))

            while not chosen_rois:
                print(f'\n--- {func_name} creation ---')
                roi_ans = input(
                    "Type a comma-separated list of the ROIs (listed above) you want to produce a figure for, "
                    "'e.g. 2, 15, 7, 23' or 'all' for all rois. \nAlternatively press enter twice to skip this step: ")

                if roi_ans.lower() == "all":
                    chosen_rois = list(range(0, len(all_rois)))

                elif len(roi_ans) > 0:
                    chosen_rois = [x.strip() for x in roi_ans.split(',')]  # Split by comma and whitespace

                    try:
                        chosen_rois = list(map(int, chosen_rois))  # Convert each list item to integers
                    except ValueError:
                        print('Comma-separated list contains non integers.\n')
                        chosen_rois = []

                else:  # Else statement for blank input, this skips creating this plot
                    chosen_rois = []
                    break

        else:  # Else if an ROI selection has been made, convert it into the correct format
            if isinstance(config_region_var, list) and isinstance(config_region_var[0], str) \
                    and config_region_var[0].lower() == "all":
                chosen_rois = list(range(0, len(all_rois)))
            else:
                chosen_rois = config_region_var

                if isinstance(chosen_rois, int):
                    chosen_rois = [chosen_rois]
                else:
                    chosen_rois = list(chosen_rois)

        chosen_rois = [all_rois[roi_number] for roi_number in chosen_rois]

        return chosen_rois

    @staticmethod
    def read_combined_results(folder, averaging_type):
        df = pd.DataFrame()
        path = ''

        try:
            if averaging_type == 'Session averaged':
                path = f"{folder}/Overall/Summarised_results/Session_averaged_results/combined_results.json"
                df = pd.read_json(path)

            elif averaging_type == 'Participant averaged':
                path = f"{folder}/Overall/Summarised_results/Participant_averaged_results/combined_results.json"
                df = pd.read_json(path)

        except ValueError:
            raise Exception(f"combined_results.json not found in {config.averaging_type} folder.")

        df = df.sort_values('index')

        return df, path

    @staticmethod
    def find_column_locs(table):
        table.columns = [x.lower() for x in table.columns]  # Convert to lower case for comparison to key later

        ignore_column_loc = next((counter for counter, column in enumerate(table.columns) if "ignore file" in column),
                                 False)

        baseline_column_loc = next((counter for counter, column in enumerate(table.columns)
                                    if "baseline" in column), False)
        if not baseline_column_loc:
            raise NameError(f'No baseline column found in {config.parameter_file}.')

        critical_column_locs = []
        for key in config.parameter_dict:
            column_loc = next((counter for counter, column in enumerate(table.columns) if key.lower() == column), False)

            if column_loc:
                critical_column_locs.append(column_loc)
            else:
                raise Exception(f'Key "{key}" not found in {config.parameter_file}. Check the Critical Parameters '
                                f'option in the Parsing menu (parameter_dict1 if not using the GUI) correctly match '
                                f'the {config.parameter_file} headers.')

        return ignore_column_loc, critical_column_locs, baseline_column_loc

    @staticmethod
    def load_paramValues_file(directory=''):
        if directory:
            folder = directory
        else:
            folder = os.getcwd()

        if os.path.isfile(f"{folder}/{config.parameter_file}"):
            table = pd.read_csv(f"{folder}/{config.parameter_file}")  # Load param table
            folder_type = 'base_folder'

        else:
            try:
                table = pd.read_csv(f"{folder}/copy_paramValues.csv")  # Load param table
                folder_type = 'report_folder'

            except FileNotFoundError:
                raise Exception('Make sure a copy of paramValues.csv is in the chosen folder. \n'
                                'Also make sure the selected folder contains all the participant directories '
                                'in the necessary BIDS format e.g. sub-01.')

        return table, folder_type

    @staticmethod
    def mk_dir(path):
        try:
            os.mkdir(path)
        except FileExistsError:
            pass

    @staticmethod
    def instance_method_handler(*argv):
        return getattr(argv[0], argv[1])(*argv[2:])

    @staticmethod
    def class_method_handler(*argv):
        return argv[0](*argv[1:])

    @staticmethod
    def start_processing_pool(restart=False):
        if config.max_core_usage == 'max':
            workers = mp.cpu_count()
        else:
            workers = int(config.max_core_usage)

        ctx = mp.get_context('forkserver')  # This stops segmentation fault for MacOS

        if config.verbose and not restart:
            print(f"\nStarting processing pool using {workers} cores.")

        return ctx.Pool(processes=workers)

    @staticmethod
    def join_processing_pool(pool, restart):
        pool.close()
        pool.join()

        if restart:
            pool = Utils.start_processing_pool(restart=True)

        return pool

    @classmethod
    def load_config(cls, config_path, filename, save=True, path=None):
        with open(f'{config_path}/{filename}', 'r') as tomlfile:
            try:
                parse = tomlfile.readlines()
                parse = toml.loads(''.join(parse))

                for key in parse:
                    if parse[key] == 'None':
                        parse[key] = None

                if save:
                    global config

                config = SimpleNamespace(**parse)

                if os.path.split(config_path)[-1] == 'roi_analysis':
                    config.statistic_options = {
                        'Participant averaged': ['Total voxels',
                                                 'Excluded voxels',
                                                 'Average voxels per session',
                                                 'Mean',
                                                 'Std_dev',
                                                 'Confidence_interval',
                                                 'Median',
                                                 'Minimum',
                                                 'Maximum',
                                                 'Participants',
                                                 'Sessions'],

                        'Session averaged': ['Voxel_amount',
                                             'Excluded_voxels_amount',
                                             'Average_voxels',
                                             'Mean',
                                             'Standard_deviation',
                                             'Confidence_interval',
                                             'Median',
                                             'Minimum',
                                             'Maximum',
                                             'Sessions']
                    }

                    config.parameter_dict = {config.parameter_dict1[i]:
                                                 config.parameter_dict2[i] for i in range(len(config.parameter_dict1))}

                    Utils.clean_config_options(config)

                if filename == 'test_config.toml':
                    # Set test config settings so files are retrieved from and outputted to the correct place
                    config.brain_file_loc = path
                    config.base_folder = path
                    config.output_folder = 'test_ROI_report'
                    config.output_folder_name = 'test_maps'
                    config.stat_map_folder = 'test_maps'

                return config

            except (toml.decoder.TomlDecodeError, AttributeError):
                raise Exception('Config file not in correct format or missing entries.')

    @staticmethod
    def clean_config_options(config):
        atlas_options = ['Cerebellum-MNIflirt', 'Cerebellum-MNIfnirt', 'HarvardOxford-cort',
                         'HarvardOxford-sub', 'JHU-ICBM-labels', 'JHU-ICBM-tracts', 'juelich', 'MNI',
                         'SMATT-labels', 'STN',
                         'striatum-structural', 'Talairach-labels', 'Thalamus']

        if not type(config.atlas_number) is int:
            config.atlas_number = atlas_options.index(config.atlas_number)

        conf_level_options = ['80%, 1.28', '85%, 1.44', '90%, 1.64', '95%, 1.96', '98%, 2.33', '99%, 2.58']

        if not type(config.conf_level_number) is int:
            config.bootstrap_alpha = 1 - float(f"0.{re.split('%', config.conf_level_number)[0]}")
            config.conf_level_number = conf_level_options.index(config.conf_level_number)

        return config

    @staticmethod
    def load_brain(file_path):
        data = nib.load(file_path)

        header = data.header
        data = data.get_fdata()

        return data, header

    @staticmethod
    def strip_ext(path):
        extensions = ['.nii.gz', '.nii', '.hdr', '.json']

        for extension in extensions:
            path = path.replace(extension, "")

        return path

    @staticmethod
    def find_participant_dirs(directory):
        # Searches for folders that start with e.g. sub-01
        participant_paths = [direc for direc in glob(f"{directory}/*") if re.search("sub-[0-9]+$", direc)]
        participant_names = [participant.split('/')[-1] for participant in participant_paths]

        if len(participant_paths) == 0:
            raise FileNotFoundError('Participant directories not found.\n'
                                    'Make sure participant directories are labelled e.g. sub-01 and the selected '
                                    'directory contains all participant directories.')
        elif config.verbose:
            print(f'Found {len(participant_paths)} participant folders.')

        return participant_paths, participant_names

    @staticmethod
    def checkversion(frat_version):
        global version
        version = frat_version

        # Check Python version:
        expect_major = 3
        expect_minor = 10

        print(f"\nfRAT {frat_version} is developed and tested with Python {str(expect_major)}.{str(expect_minor)}.\n")
        if sys.version_info[:2] != (expect_major, expect_minor):
            current_version = f"{str(sys.version_info[0])}.{str(sys.version_info[1])}.{str(sys.version_info[2])}"
            print(f"INFO: Python version {current_version} is untested. Consider changing to version "
                  f"{str(expect_major)}.{str(expect_minor)} if there are errors running fRAT.")

    @staticmethod
    def chdir_to_output_directory(current_step, config):
        if current_step == 'Statistics' and config.run_plotting:
            return  # Will already be in the correct directory

        elif config.run_analysis:
            from .analysis import Environment_Setup
            json_directory = f'{os.getcwd()}/{Environment_Setup.save_location}'

            os.chdir(json_directory)

        elif config.report_output_folder in ("", " "):
            print('Select the directory output by the fRAT.')
            json_directory = Utils.file_browser(title='Select the directory output by the fRAT', chdir=True)

        else:
            json_directory = config.report_output_folder

            try:
                os.chdir(json_directory)
            except FileNotFoundError:
                raise FileNotFoundError(
                    'Output folder location (fRAT output folder location) is not a valid directory.')

            if config.verbose:
                print(f'Output folder selection: {json_directory}')

        return json_directory



import itertools

import numpy as np
from nipype.interfaces import fsl
import logging
import time

from .utils import *

config = None
config_path = ''
config_filename = ''


def file_setup(func):
    file_location = config.input_folder_name

    if config.output_folder_name != 'DEFAULT':
        output_folder = f"statmaps/{config.output_folder_name}"
    elif func == 'Image SNR':
        output_folder = 'statmaps/imageSNR_report'
    elif func == 'Temporal SNR':
        output_folder = 'statmaps/temporalSNR_report'
    elif func == 'Add Gaussian noise':
        output_folder = 'added_noise'

    if config.base_folder in ("", " "):
        print('Select the directory which contains the subject folders.')
        base_sub_location = Utils.file_browser(title='Select the directory which contains the subject folders')

    else:
        base_sub_location = config.base_folder

    # Create dictionary from each participant directory
    participant_dir, _ = Utils.find_participant_dirs(base_sub_location)
    participants = {participant: None for participant in participant_dir}

    for participant in participants:
        # Find all nifti and analyze files
        participants[participant] = Utils.find_files(f"{participant}/{file_location}", "hdr", "nii.gz", "nii")

        if 'statmaps' in output_folder:
            Utils.check_and_make_dir(f"{participant}/statmaps")
            Utils.check_and_make_dir(f"{participant}/{output_folder}", delete_old=True)
            Utils.save_config(f"{participant}/{output_folder}", config_path, config_filename,
                              additional_info=[f"statistical_map_created = '{func}'\n"],
                              new_config_name='statmap_config')
        else:
            Utils.check_and_make_dir(f"{participant}/{file_location}/{output_folder}", delete_old=True)

    return participants, output_folder, file_location


def calculate_sigma_in_volumes(file_path):
    data = nib.load(file_path)
    TR = data.header['pixdim'][4]  # Find TR

    # Equation found here: https://www.jiscmail.ac.uk/cgi-bin/webadmin?A2=FSL;f6fd75a6.1709
    return 1 / (2 * config.highpass_filter_cutoff * TR)


def save_brain(data, ext, no_ext_file, output_folder, header=None):
    brain = nib.Nifti1Pair(data, None, header)
    nib.save(brain, f"{output_folder}/{no_ext_file}{ext}.nii.gz")

    return f"{output_folder}/{no_ext_file}{ext}.nii.gz"


def temporalSNR_calc(file, no_ext_file, output_folder):
    fsl.maths.MeanImage(in_file=file, out_file=f'{output_folder}/{no_ext_file}_tMean.nii.gz').run()  # Mean over time

    fsl.maths.StdImage(in_file=file, out_file=f'{output_folder}/{no_ext_file}_tStd.nii.gz').run()  # Standard dev over time

    # tMean / tStd
    fsl.maths.BinaryMaths(operation='div', in_file=f'{output_folder}/{no_ext_file}_tMean.nii.gz',
                      operand_file=f'{output_folder}/{no_ext_file}_tStd.nii.gz',
                      out_file=f'{output_folder}/{no_ext_file}_tSNR.nii.gz').run()

    # Threshold volume so any tSNR values above 1000 are set to 0
    fsl.maths.Threshold(in_file=f'{output_folder}/{no_ext_file}_tSNR.nii.gz', thresh=1000.0, direction='above',
                    out_file=f'{output_folder}/{no_ext_file}_tSNR.nii.gz').run()


def imageSNR_calc(func_file, noise_file, no_ext_file, output_folder, participant_dir):
    fsl.maths.MeanImage(in_file=func_file, out_file=f'{output_folder}/{no_ext_file}_tMean.nii.gz').run()  # Mean over time

    if config.iSNR_std_use_only_nonzero_voxels:
        std_string = '-S'
    else:
        std_string = '-s'

    if config.noise_volume:
        std = fsl.ImageStats(in_file=noise_file, op_string=std_string,
                         terminal_output='allatonce').run()  # Std dev of entire volume
        noise_value = std.outputs.get()['out_stat']

    else:
        base_sub_location = Path(participant_dir).parents[0]
        participant_name = os.path.split(Path(participant_dir))[1]

        noise_value_csv = pd.read_csv(f'{base_sub_location}/noiseValues.csv')
        noise_value = float(noise_value_csv[noise_value_csv['Participant'] == participant_name]['Background noise'])

    # tMean / Std
    fsl.maths.BinaryMaths(operation='div',
                      in_file=f'{output_folder}/{no_ext_file}_tMean.nii.gz',
                      operand_value=noise_value,
                      out_file=f'{output_folder}/{no_ext_file}_iSNR.nii.gz').run()

    if config.magnitude_correction:
        magnitude_correction(f'{output_folder}/{no_ext_file}_iSNR.nii.gz')


def magnitude_correction(file_name):
    fsl.maths.BinaryMaths(in_file=file_name, operation='mul', operand_value=0.7, out_file=file_name).run()


def separate_noise_from_func(file, no_ext_file, output_folder, participant):
    data, header = Utils.load_brain(file)

    if config.noise_volume_location == 'End':
        noise_data, func_data = data[:, :, :, -1], data[:, :, :, :-1]
    elif config.noise_volume_location == 'Beginning':
        noise_data, func_data = data[:, :, :, 0], data[:, :, :, 1:]
    else:
        raise Exception('Noise volume location not valid.')

    noise_file = save_brain(noise_data, '_noise_volume', no_ext_file, output_folder, header)

    func_file = save_to_cleaned_folder(func_data, header, no_ext_file, participant,
                                       f'Removed noise volume from {config.noise_volume_location.lower()} of timeseries')

    return func_file, noise_file


def save_to_cleaned_folder(data, header, no_ext_file, participant, method):
    with open(f'{participant}/{config.input_folder_name}_cleaned/changes_made_to_files.txt', 'a+') as f:
        f.seek(0)  # Go to start of file
        f.write(f'{no_ext_file}: {method}\n')

    data = save_brain(data, '', no_ext_file, f'{participant}/{config.input_folder_name}_cleaned', header)

    return data


def highpass_filtering(file_path, output_folder, no_ext_file):
    sigma_in_volumes = calculate_sigma_in_volumes(file_path)

    fsl.maths.MeanImage(in_file=f'{file_path}', out_file=f'{output_folder}/{no_ext_file}_tMean.nii.gz').run()

    fsl.TemporalFilter(in_file=f'{file_path}',
                   out_file=f'{output_folder}/{no_ext_file}_filtered.nii.gz',
                   highpass_sigma=sigma_in_volumes).run()

    fsl.maths.BinaryMaths(operation='add', in_file=f'{output_folder}/{no_ext_file}_tMean.nii.gz',
                      operand_file=f'{output_folder}/{no_ext_file}_filtered.nii.gz',
                      out_file=f'{output_folder}/{no_ext_file}_filtered_restoredmean.nii.gz').run()

    return f'{output_folder}/{no_ext_file}_filtered_restoredmean.nii.gz', \
           f'{output_folder}/{no_ext_file}_filtered.nii.gz'


def delete_files(redundant_files):
    for file in redundant_files:
        os.remove(file)


def create_statmaps(func, file, no_ext_file, noise_file, output_folder, participant):
    if func == 'Image SNR':
        imageSNR_calc(file, noise_file, no_ext_file, output_folder, participant)
    elif func == 'Temporal SNR':
        temporalSNR_calc(file, no_ext_file, output_folder)


def prepare_statmap_files(file, no_ext_file, output_folder, participant):
    noise_file = None

    redundant_files = []
    outliers = []

    # Save original copy of file which may be overwritten later, however allows this folder to always be used
    data, header = Utils.load_brain(file)
    save_brain(data, '', no_ext_file, f'{participant}/{config.input_folder_name}_cleaned', header)

    if config.noise_volume:
        file, noise_file = separate_noise_from_func(file, no_ext_file, output_folder, participant)
        redundant_files.append(noise_file)

    if config.remove_motion_outliers:
        file, outlier_timepoints, outlier_files = remove_motion_outliers(file, no_ext_file, output_folder, participant)
        redundant_files.extend(outlier_files)
        outliers.extend(outlier_timepoints)

    if config.motion_correction:
        output = f'{output_folder}/{no_ext_file}_motion_corrected.nii.gz'

        fsl.MCFLIRT(in_file=file, out_file=output).run()
        file = output

        data, header = Utils.load_brain(file)
        file = save_to_cleaned_folder(data, header, no_ext_file, participant, 'Motion corrected data')

    if config.spatial_smoothing:
        fsl.SUSAN(in_file=file, fwhm=config.smoothing_fwhm, brightness_threshold=config.smoothing_brightness_threshold,
                  out_file=f'{output_folder}/{no_ext_file}_smoothed.nii.gz').run()
        file = f'{output_folder}/{no_ext_file}_smoothed.nii.gz'

    if config.temporal_filter:
        file, redundant_file = highpass_filtering(file, output_folder, no_ext_file)
        redundant_files.extend([file, redundant_file])

    return file, noise_file, redundant_files, outliers


def remove_motion_outliers(file, no_ext_file, output_folder, participant):
    outlier_file = f'{output_folder}/{no_ext_file}_outliers.txt'
    outlier_plot = f'{output_folder}/{no_ext_file}_metrics.png'
    outlier_values = f'{output_folder}/{no_ext_file}_metrics.txt'

    fsl.MotionOutliers(in_file=file, out_file=outlier_file,
                       out_metric_plot=outlier_plot,
                       out_metric_values=outlier_values).run()

    with open(f'{output_folder}/{no_ext_file}_outliers.txt') as f:
        lines = f.readlines()

    outlier_timepoints = []
    for time_point, line in enumerate(lines):
        outlier_check = line.replace(" ", "").find('1')
        if outlier_check != -1:
            outlier_timepoints.append(time_point)

    data, header = Utils.load_brain(file)
    data = np.delete(data, outlier_timepoints, axis=3)

    # Save copy of motion outlier removed files as it may make registration better during main analysis
    file = save_to_cleaned_folder(data, header, no_ext_file, participant, 'Removed motion outlier timepoints')

    return file, [no_ext_file, len(outlier_timepoints)], [outlier_file, outlier_plot, outlier_values]


def process_files_for_statmaps(file, participant, output_folder, file_location, func, cfg):
    global config
    config = cfg

    no_ext_file = Utils.strip_ext(file)
    file = f"{participant}/{file_location}/{file}"
    output_folder = f'{participant}/{output_folder}'

    if config.verbose:
        print(f'        Analysing file: {no_ext_file}')

    file, noise_file, redundant_files, outliers = prepare_statmap_files(file, no_ext_file, output_folder, participant)
    create_statmaps(func, file, no_ext_file, noise_file, output_folder, participant)

    delete_files(redundant_files)

    return outliers


def calculate_statistical_maps(participants, output_folder, file_location, func):
    if config.multicore_processing:
        pool = Utils.start_processing_pool()
    else:
        pool = None

    if config.verbose:
        print(f'\nSearching in folder: {config.input_folder_name}'
              f'\nSaving output in directory: {output_folder}')

    for participant_dir, files in participants.items():
        if config.verbose:
            print(f'\nCreating statistical maps for participant: {participant_dir.split("/")[-1]}'
                  f'\n      Creating statmaps for {len(files)} files')

        Utils.check_and_make_dir(f'{participant_dir}/{config.input_folder_name}_cleaned', delete_old=True)

        iterable = zip(files,
                       itertools.repeat(participant_dir),
                       itertools.repeat(output_folder),
                       itertools.repeat(file_location),
                       itertools.repeat(func),
                       itertools.repeat(config))

        if config.multicore_processing:
            outliers = list(pool.starmap(process_files_for_statmaps, iterable))
        else:
            outliers = list(itertools.starmap(process_files_for_statmaps, iterable))

        if config.remove_motion_outliers:
            outliers = pd.DataFrame(columns=['File', 'Outliers removed'], data=outliers).sort_values('Outliers removed')

            with open(f"{participant_dir}/{output_folder}/number_of_outliers_removed.csv", "w") as f:
                f.write(outliers.to_csv(index=False))

    if config.multicore_processing:
        Utils.join_processing_pool(pool, restart=False)


def prepare_to_run_utility(participants, output_folder, file_location, func):
    if config.multicore_processing:
        pool = Utils.start_processing_pool()
    else:
        pool = None

    if config.verbose:
        print(f'\nSaving output in directory: {config.input_folder_name}/{output_folder}')

    for participant, files in participants.items():
        participant_dir = f"{participant}/{file_location}"

        if config.verbose:
            print(f"\nRunning '{func}' utility on participant: {participant_dir.split('/')[-2]}"
                  f"\n      Running utility on {len(files)} files")

        iterable = zip(files,
                       itertools.repeat(participant_dir),
                       itertools.repeat(output_folder),
                       itertools.repeat(func),
                       itertools.repeat(config))

        if config.multicore_processing:
            list(pool.starmap(run_utility, iterable))

        else:
            list(itertools.starmap(run_utility, iterable))

    if config.multicore_processing:
        Utils.join_processing_pool(pool, restart=False)


def run_utility(file, participant_dir, output_folder, func, cfg):
    global config
    config = cfg

    no_ext_file = Utils.strip_ext(file)
    file = f"{participant_dir}/{file}"
    base_sub_location = Path(participant_dir).parents[1]
    participant_name = os.path.split(Path(participant_dir).parents[0])[1]

    if config.verbose:
        print(f'        Running utility on file: {no_ext_file}')

    if func == 'Add Gaussian noise':
        noise_value_csv = pd.read_csv(f'{base_sub_location}/noiseValues.csv')
        participant_noise_level = float(
            noise_value_csv[noise_value_csv['Participant'] == participant_name]['Noise over time'])
        add_noise_to_file(file, no_ext_file, participant_dir, output_folder, participant_noise_level)


def add_noise_to_file(file, no_ext_file, participant_dir, output_folder, participant_noise_level):
    data, header = Utils.load_brain(file)

    # Save initial data with no added noise
    save_brain(data, ext=f'_noiselevel0', no_ext_file=no_ext_file,
               output_folder=f"{participant_dir}/{output_folder}", header=header)

    for multiplier in config.noise_multipliers:
        gaussian_noise = np.random.default_rng().normal(loc=0.0,
                                                        scale=participant_noise_level * multiplier,
                                                        size=data.shape)

        noisy_data = data + gaussian_noise
        noisy_data[noisy_data < 0] = 0  # Set values below 0 to 0

        save_brain(noisy_data, ext=f'_noiselevel{multiplier}', no_ext_file=no_ext_file,
                   output_folder=f"{participant_dir}/{output_folder}", header=header)


def main(func, version, config_file, path=None):
    global config, config_path, config_filename

    start_time = time.time()
    Utils.checkversion(version)

    # Set global variables
    config_path = f'{Path(os.path.abspath(__file__)).parents[1]}/configuration_profiles/maps/'
    config_filename = config_file

    # Load config file
    config = Utils.load_config(config_path, config_file, path=path)

    if config.verbose:
        print('\n--------------------------------\n'
              '--- Statistical map creation ---\n'
              '--------------------------------\n'
              f'\nCreating {func} maps.\n')

    logging.getLogger('nipype.workflow').setLevel(0)  # Suppress workflow terminal output

    if func == 'Image SNR' and not config.noise_volume and config.verbose:
        print('"Noise volume included in time series" is false. Trying to find values for each participant in '
              'noiseValues.csv instead. If this is not correct, set "Noise volume included in time series" to true.\n')

    participants, output_folder, file_location = file_setup(func)

    if func in ['Add Gaussian noise']:
        prepare_to_run_utility(participants, output_folder, file_location, func)

    else:
        calculate_statistical_maps(participants, output_folder, file_location, func)

    if config.verbose:
        print(f"\n--- Completed in {round((time.time() - start_time), 2)} seconds ---\n\n")
